@@ -1,22 +1,16 @@
-# -*- coding: utf-8 -*-
-from __future__ import unicode_literals
-
 import functools
 import inspect
+import typing
 import logging
 import warnings
-from copy import copy
-
-try:
-    import simplejson as json
-except ImportError:
-    import json  # type: ignore
+from collections.abc import Mapping
+import json
 
 import marshmallow as ma
 from marshmallow import ValidationError
-from marshmallow.utils import missing, is_collection
+from marshmallow.utils import missing
 
-from webargs.compat import Mapping, iteritems, MARSHMALLOW_VERSION_INFO
+from webargs.compat import MARSHMALLOW_VERSION_INFO
 from webargs.dict2schema import dict2schema
 from webargs.fields import DelimitedList
 
@@ -28,7 +22,6 @@ __all__ = [
     "dict2schema",
     "is_multiple",
     "Parser",
-    "get_value",
     "missing",
     "parse_json",
 ]
@@ -42,9 +35,8 @@ def _callable_or_raise(obj):
     callable, a ValueError is raised.
     """
     if obj and not callable(obj):
-        raise ValueError("{0!r} is not callable.".format(obj))
-    else:
-        return obj
+        raise ValueError("{!r} is not callable.".format(obj))
+    return obj
 
 
 def is_multiple(field):
@@ -74,53 +66,17 @@ def is_json(mimetype):
     return False
 
 
-def get_value(data, name, field, allow_many_nested=False):
-    """Get a value from a dictionary. Handles ``MultiDict`` types when
-    ``field`` handles repeated/multi-value arguments.
-    If the value is not found, return `missing`.
-
-    :param object data: Mapping (e.g. `dict`) or list-like instance to
-        pull the value from.
-    :param str name: Name of the key.
-    :param bool allow_many_nested: Whether to allow a list of nested objects
-        (it is valid only for JSON format, so it is set to True in ``parse_json``
-        methods).
-    """
-    missing_value = missing
-    if allow_many_nested and isinstance(field, ma.fields.Nested) and field.many:
-        if is_collection(data):
-            return data
-
-    if not hasattr(data, "get"):
-        return missing_value
-
-    multiple = is_multiple(field)
-    val = data.get(name, missing_value)
-    if multiple and val is not missing:
-        if hasattr(data, "getlist"):
-            return data.getlist(name)
-        elif hasattr(data, "getall"):
-            return data.getall(name)
-        elif isinstance(val, (list, tuple)):
-            return val
-        if val is None:
-            return None
-        else:
-            return [val]
-    return val
-
-
-def parse_json(s, encoding="utf-8"):
-    if isinstance(s, bytes):
+def parse_json(string, *, encoding="utf-8"):
+    if isinstance(string, bytes):
         try:
-            s = s.decode(encoding)
-        except UnicodeDecodeError as e:
+            string = string.decode(encoding)
+        except UnicodeDecodeError as exc:
             raise json.JSONDecodeError(
-                "Bytes decoding error : {}".format(e.reason),
-                doc=str(e.object),
-                pos=e.start,
+                "Bytes decoding error : {}".format(exc.reason),
+                doc=str(exc.object),
+                pos=exc.start,
             )
-    return json.loads(s)
+    return json.loads(string)
 
 
 def _ensure_list_of_callables(obj):
@@ -130,27 +86,26 @@ def _ensure_list_of_callables(obj):
         elif callable(obj):
             validators = [obj]
         else:
-            raise ValueError(
-                "{0!r} is not a callable or list of callables.".format(obj)
-            )
+            raise ValueError("{!r} is not a callable or list of callables.".format(obj))
     else:
         validators = []
     return validators
 
 
-class Parser(object):
+class Parser:
     """Base parser class that provides high-level implementation for parsing
     a request.
 
-    Descendant classes must provide lower-level implementations for parsing
-    different locations, e.g. ``parse_json``, ``parse_querystring``, etc.
+    Descendant classes must provide lower-level implementations for reading
+    data from  different locations, e.g. ``load_json``, ``load_querystring``,
+    etc.
 
-    :param tuple locations: Default locations to parse.
+    :param str location: Default location to use for data
     :param callable error_handler: Custom error handler function.
     """
 
-    #: Default locations to check for data
-    DEFAULT_LOCATIONS = ("querystring", "form", "json")
+    #: Default location to check for data
+    DEFAULT_LOCATION = "json"
     #: The marshmallow Schema class to use when creating new schemas
     DEFAULT_SCHEMA_CLASS = ma.Schema
     #: Default status code to return for validation errors
@@ -160,38 +115,31 @@ class Parser(object):
 
     #: Maps location => method name
     __location_map__ = {
-        "json": "parse_json",
-        "querystring": "parse_querystring",
-        "query": "parse_querystring",
-        "form": "parse_form",
-        "headers": "parse_headers",
-        "cookies": "parse_cookies",
-        "files": "parse_files",
+        "json": "load_json",
+        "querystring": "load_querystring",
+        "query": "load_querystring",
+        "form": "load_form",
+        "headers": "load_headers",
+        "cookies": "load_cookies",
+        "files": "load_files",
+        "json_or_form": "load_json_or_form",
     }
 
-    def __init__(self, locations=None, error_handler=None, schema_class=None):
-        self.locations = locations or self.DEFAULT_LOCATIONS
+    def __init__(self, location=None, *, error_handler=None, schema_class=None):
+        self.location = location or self.DEFAULT_LOCATION
         self.error_callback = _callable_or_raise(error_handler)
         self.schema_class = schema_class or self.DEFAULT_SCHEMA_CLASS
-        #: A short-lived cache to store results from processing request bodies.
-        self._cache = {}
 
-    def _validated_locations(self, locations):
-        """Ensure that the given locations argument is valid.
+    def _get_loader(self, location):
+        """Get the loader function for the given location.
 
-        :raises: ValueError if a given locations includes an invalid location.
+        :raises: ValueError if a given location is invalid.
         """
-        # The set difference between the given locations and the available locations
-        # will be the set of invalid locations
         valid_locations = set(self.__location_map__.keys())
-        given = set(locations)
-        invalid_locations = given - valid_locations
-        if len(invalid_locations):
-            msg = "Invalid locations arguments: {0}".format(list(invalid_locations))
+        if location not in valid_locations:
+            msg = "Invalid location argument: {}".format(location)
             raise ValueError(msg)
-        return locations
 
-    def _get_handler(self, location):
         # Parsing function to call
         # May be a method name (str) or a function
         func = self.__location_map__.get(location)
@@ -201,82 +149,41 @@ class Parser(object):
             else:
                 function = getattr(self, func)
         else:
-            raise ValueError('Invalid location: "{0}"'.format(location))
+            raise ValueError('Invalid location: "{}"'.format(location))
         return function
 
-    def _get_value(self, name, argobj, req, location):
-        function = self._get_handler(location)
-        return function(req, name, argobj)
+    def _load_location_data(self, *, schema, req, location):
+        """Return a dictionary-like object for the location on the given request.
 
-    def parse_arg(self, name, field, req, locations=None):
-        """Parse a single argument from a request.
-
-        .. note::
-            This method does not perform validation on the argument.
-
-        :param str name: The name of the value.
-        :param marshmallow.fields.Field field: The marshmallow `Field` for the request
-            parameter.
-        :param req: The request object to parse.
-        :param tuple locations: The locations ('json', 'querystring', etc.) where
-            to search for the value.
-        :return: The unvalidated argument value or `missing` if the value cannot
-            be found on the request.
+        Needs to have the schema in hand in order to correctly handle loading
+        lists from multidict objects and `many=True` schemas.
         """
-        location = field.metadata.get("location")
-        if location:
-            locations_to_check = self._validated_locations([location])
-        else:
-            locations_to_check = self._validated_locations(locations or self.locations)
-
-        for location in locations_to_check:
-            value = self._get_value(name, field, req=req, location=location)
-            # Found the value; validate and return it
-            if value is not missing:
-                return value
-        return missing
-
-    def _parse_request(self, schema, req, locations):
-        """Return a parsed arguments dictionary for the current request."""
-        if schema.many:
-            assert (
-                "json" in locations
-            ), "schema.many=True is only supported for JSON location"
-            # The ad hoc Nested field is more like a workaround or a helper,
-            # and it servers its purpose fine. However, if somebody has a desire
-            # to re-design the support of bulk-type arguments, go ahead.
-            parsed = self.parse_arg(
-                name="json",
-                field=ma.fields.Nested(schema, many=True),
-                req=req,
-                locations=locations,
-            )
-            if parsed is missing:
-                parsed = []
-        else:
-            argdict = schema.fields
-            parsed = {}
-            for argname, field_obj in iteritems(argdict):
-                if MARSHMALLOW_VERSION_INFO[0] < 3:
-                    parsed_value = self.parse_arg(argname, field_obj, req, locations)
-                    # If load_from is specified on the field, try to parse from that key
-                    if parsed_value is missing and field_obj.load_from:
-                        parsed_value = self.parse_arg(
-                            field_obj.load_from, field_obj, req, locations
-                        )
-                        argname = field_obj.load_from
-                else:
-                    argname = field_obj.data_key or argname
-                    parsed_value = self.parse_arg(argname, field_obj, req, locations)
-                if parsed_value is not missing:
-                    parsed[argname] = parsed_value
-        return parsed
+        loader_func = self._get_loader(location)
+        data = loader_func(req, schema)
+        # when the desired location is empty (no data), provide an empty
+        # dict as the default so that optional arguments in a location
+        # (e.g. optional JSON body) work smoothly
+        if data is missing:
+            data = {}
+        return data
 
     def _on_validation_error(
-        self, error, req, schema, error_status_code, error_headers
+        self, error, req, schema, location, *, error_status_code, error_headers
     ):
+        # rewrite messages to be namespaced under the location which created
+        # them
+        # e.g. {"json":{"foo":["Not a valid integer."]}}
+        #      instead of
+        #      {"foo":["Not a valid integer."]}
+        error.messages = {location: error.messages}
         error_handler = self.error_callback or self.handle_error
-        error_handler(error, req, schema, error_status_code, error_headers)
+        error_handler(
+            error,
+            req,
+            schema,
+            error_status_code=error_status_code,
+            error_headers=error_headers,
+        )
 
     def _validate_arguments(self, data, validators):
         for validator in validators:
@@ -300,7 +207,7 @@ class Parser(object):
         elif callable(argmap):
             schema = argmap(req)
         else:
-            schema = dict2schema(argmap, self.schema_class)()
+            schema = dict2schema(argmap, schema_class=self.schema_class)()
         if MARSHMALLOW_VERSION_INFO[0] < 3 and not schema.strict:
             warnings.warn(
                 "It is highly recommended that you set strict=True on your schema "
@@ -309,19 +216,15 @@ class Parser(object):
             )
         return schema
 
-    def _clone(self):
-        clone = copy(self)
-        clone.clear_cache()
-        return clone
-
     def parse(
         self,
         argmap,
         req=None,
-        locations=None,
+        *,
+        location=None,
         validate=None,
         error_status_code=None,
-        error_headers=None,
+        error_headers=None
     ):
         """Main request parsing method.
 
@@ -329,9 +232,10 @@ class Parser(object):
             of argname -> `marshmallow.fields.Field` pairs, or a callable
             which accepts a request and returns a `marshmallow.Schema`.
         :param req: The request object to parse.
-        :param tuple locations: Where on the request to search for values.
-            Can include one or more of ``('json', 'querystring', 'form',
-            'headers', 'cookies', 'files')``.
+        :param str location: Where on the request to load values.
+            Can be any of the values in :py:attr:`~__location_map__`. By
+            default, that means one of ``('json', 'query', 'querystring',
+            'form', 'headers', 'cookies', 'files', 'json_or_form')``.
         :param callable validate: Validation function or list of validation functions
             that receives the dictionary of parsed arguments. Validator either returns a
             boolean or raises a :exc:`ValidationError`.
@@ -342,37 +246,31 @@ class Parser(object):
 
          :return: A dictionary of parsed arguments
         """
-        self.clear_cache()  # in case someone used `parse_*()`
         req = req if req is not None else self.get_default_request()
-        assert req is not None, "Must pass req object"
+        location = location or self.location
+        if req is None:
+            raise ValueError("Must pass req object")
         data = None
         validators = _ensure_list_of_callables(validate)
-        parser = self._clone()
         schema = self._get_schema(argmap, req)
         try:
-            parsed = parser._parse_request(
-                schema=schema, req=req, locations=locations or self.locations
+            location_data = self._load_location_data(
+                schema=schema, req=req, location=location
             )
-            result = schema.load(parsed)
+            result = schema.load(location_data)
             data = result.data if MARSHMALLOW_VERSION_INFO[0] < 3 else result
-            parser._validate_arguments(data, validators)
+            self._validate_arguments(data, validators)
         except ma.exceptions.ValidationError as error:
-            parser._on_validation_error(
-                error, req, schema, error_status_code, error_headers
+            self._on_validation_error(
+                error,
+                req,
+                schema,
+                location,
+                error_status_code=error_status_code,
+                error_headers=error_headers,
             )
         return data
 
-    def clear_cache(self):
-        """Invalidate the parser's cache.
-
-        This is usually a no-op now since the Parser clone used for parsing a
-        request is discarded afterwards.  It can still be used when manually
-        calling ``parse_*`` methods which would populate the cache on the main
-        Parser instance.
-        """
-        self._cache = {}
-        return None
-
     def get_default_request(self):
         """Optional override. Provides a hook for frameworks that use thread-local
         request objects.
@@ -393,29 +291,40 @@ class Parser(object):
         """
         return None
 
+    @staticmethod
+    def _update_args_kwargs(args, kwargs, parsed_args, as_kwargs):
+        """Update args or kwargs with parsed_args depending on as_kwargs"""
+        if as_kwargs:
+            kwargs.update(parsed_args)
+        else:
+            # Add parsed_args after other positional arguments
+            args += (parsed_args,)
+        return args, kwargs
+
     def use_args(
         self,
         argmap,
         req=None,
-        locations=None,
+        *,
+        location=None,
         as_kwargs=False,
         validate=None,
         error_status_code=None,
-        error_headers=None,
+        error_headers=None
     ):
         """Decorator that injects parsed arguments into a view function or method.
 
         Example usage with Flask: ::
 
             @app.route('/echo', methods=['get', 'post'])
-            @parser.use_args({'name': fields.Str()})
+            @parser.use_args({'name': fields.Str()}, location="querystring")
             def greet(args):
                 return 'Hello ' + args['name']
 
         :param argmap: Either a `marshmallow.Schema`, a `dict`
             of argname -> `marshmallow.fields.Field` pairs, or a callable
             which accepts a request and returns a `marshmallow.Schema`.
-        :param tuple locations: Where on the request to search for values.
+        :param str locations: Where on the request to load values.
         :param bool as_kwargs: Whether to insert arguments as keyword arguments.
         :param callable validate: Validation function that receives the dictionary
             of parsed arguments. If the function returns ``False``, the parser
@@ -425,12 +334,12 @@ class Parser(object):
         :param dict error_headers: Headers passed to error handler functions when a
             a `ValidationError` is raised.
         """
-        locations = locations or self.locations
+        location = location or self.location
         request_obj = req
         # Optimization: If argmap is passed as a dictionary, we only need
         # to generate a Schema once
         if isinstance(argmap, Mapping):
-            argmap = dict2schema(argmap, self.schema_class)()
+            argmap = dict2schema(argmap, schema_class=self.schema_class)()
 
         def decorator(func):
             req_ = request_obj
@@ -441,29 +350,27 @@ class Parser(object):
 
                 if not req_obj:
                     req_obj = self.get_request_from_view_args(func, args, kwargs)
+
                 # NOTE: At this point, argmap may be a Schema, or a callable
                 parsed_args = self.parse(
                     argmap,
                     req=req_obj,
-                    locations=locations,
+                    location=location,
                     validate=validate,
                     error_status_code=error_status_code,
                     error_headers=error_headers,
                 )
-                if as_kwargs:
-                    kwargs.update(parsed_args)
-                    return func(*args, **kwargs)
-                else:
-                    # Add parsed_args after other positional arguments
-                    new_args = args + (parsed_args,)
-                    return func(*new_args, **kwargs)
+                args, kwargs = self._update_args_kwargs(
+                    args, kwargs, parsed_args, as_kwargs
+                )
+                return func(*args, **kwargs)
 
             wrapper.__wrapped__ = func
             return wrapper
 
         return decorator
 
-    def use_kwargs(self, *args, **kwargs):
+    def use_kwargs(self, *args, **kwargs) -> typing.Callable:
         """Decorator that injects parsed arguments into a view function or method
         as keyword arguments.
 
@@ -481,19 +388,23 @@ class Parser(object):
         kwargs["as_kwargs"] = True
         return self.use_args(*args, **kwargs)
 
-    def location_handler(self, name):
-        """Decorator that registers a function for parsing a request location.
-        The wrapped function receives a request, the name of the argument, and
-        the corresponding `Field <marshmallow.fields.Field>` object.
+    def location_loader(self, name):
+        """Decorator that registers a function for loading a request location.
+        The wrapped function receives a schema and a request.
+
+        The schema will usually not be relevant, but it's important in some
+        cases -- most notably in order to correctly load multidict values into
+        list fields. Without the schema, there would be no way to know whether
+        to simply `.get()` or `.getall()` from a multidict for a given value.
 
         Example: ::
 
             from webargs import core
             parser = core.Parser()
 
-            @parser.location_handler("name")
-            def parse_data(request, name, field):
-                return request.data.get(name)
+            @parser.location_loader("name")
+            def load_data(request, schema):
+                return request.data
 
         :param str name: The name of the location to register.
         """
@@ -523,7 +434,7 @@ class Parser(object):
 
 
             @parser.error_handler
-            def handle_error(error, req, schema, status_code, headers):
+            def handle_error(error, req, schema, *, status_code, headers):
                 raise CustomError(error.messages)
 
         :param callable func: The error callback to register.
@@ -531,47 +442,94 @@ class Parser(object):
         self.error_callback = func
         return func
 
+    def _handle_invalid_json_error(self, error, req, *args, **kwargs):
+        """Internal hook for overriding treatment of JSONDecodeErrors.
+
+        Invoked by default `load_json` implementation.
+
+        External parsers can just implement their own behavior for load_json ,
+        so this is not part of the public parser API.
+        """
+        raise error
+
+    def load_json(self, req, schema):
+        """Load JSON from a request object or return `missing` if no value can
+        be found.
+        """
+        # NOTE: although this implementation is real/concrete and used by
+        # several of the parsers in webargs, it relies on the internal hooks
+        # `_handle_invalid_json_error` and `_raw_load_json`
+        # these methods are not part of the public API and are used to simplify
+        # code sharing amongst the built-in webargs parsers
+        try:
+            return self._raw_load_json(req)
+        except json.JSONDecodeError as exc:
+            if exc.doc == "":
+                return missing
+            return self._handle_invalid_json_error(exc, req)
+        except UnicodeDecodeError as exc:
+            return self._handle_invalid_json_error(exc, req)
+
+    def load_json_or_form(self, req, schema):
+        """Load data from a request, accepting either JSON or form-encoded
+        data.
+
+        The data will first be loaded as JSON, and, if that fails, it will be
+        loaded as a form post.
+        """
+        data = self.load_json(req, schema)
+        if data is not missing:
+            return data
+        return self.load_form(req, schema)
+
     # Abstract Methods
 
-    def parse_json(self, req, name, arg):
-        """Pull a JSON value from a request object or return `missing` if the
-        value cannot be found.
+    def _raw_load_json(self, req):
+        """Internal hook method for implementing load_json()
+
+        Get a request body for feeding in to `load_json`, and parse it either
+        using core.parse_json() or similar utilities which raise
+        JSONDecodeErrors.
+        Ensure consistent behavior when encountering decoding errors.
+
+        The default implementation here simply returns `missing`, and the default
+        implementation of `load_json` above will pass that value through.
+        However, by implementing a "mostly concrete" version of load_json with
+        this as a hook for getting data, we consolidate the logic for handling
+        those JSONDecodeErrors.
         """
         return missing
 
-    def parse_querystring(self, req, name, arg):
-        """Pull a value from the query string of a request object or return `missing` if
-        the value cannot be found.
+    def load_querystring(self, req, schema):
+        """Load the query string of a request object or return `missing` if no
+        value can be found.
         """
         return missing
 
-    def parse_form(self, req, name, arg):
-        """Pull a value from the form data of a request object or return
-        `missing` if the value cannot be found.
+    def load_form(self, req, schema):
+        """Load the form data of a request object or return `missing` if no
+        value can be found.
         """
         return missing
 
-    def parse_headers(self, req, name, arg):
-        """Pull a value from the headers or return `missing` if the value
-        cannot be found.
+    def load_headers(self, req, schema):
+        """Load the headers or return `missing` if no value can be found.
         """
         return missing
 
-    def parse_cookies(self, req, name, arg):
-        """Pull a cookie value from the request or return `missing` if the value
-        cannot be found.
+    def load_cookies(self, req, schema):
+        """Load the cookies from the request or return `missing` if no value
+        can be found.
         """
         return missing
 
-    def parse_files(self, req, name, arg):
-        """Pull a file from the request or return `missing` if the value file
-        cannot be found.
+    def load_files(self, req, schema):
+        """Load files from the request or return `missing` if no values can be
+        found.
         """
         return missing
 
-    def handle_error(
-        self, error, req, schema, error_status_code=None, error_headers=None
-    ):
+    def handle_error(self, error, req, schema, *, error_status_code, error_headers):
         """Called if an error occurs while parsing args. By default, just logs and
         raises ``error``.
         """
